@@ -22,15 +22,11 @@
 #include "types.h"
 #include "list.h"
 
-#define ACTOR_RESOLUTION	250	/* in millis */
-#define ACTOR_MAX_LOOPS		1
-
 typedef enum actor_state_e {
     ACTOR_INVALID,
     ACTOR_WAITING,
     ACTOR_SCHEDULED,
     ACTOR_NOTSCHEDULED,
-    ACTOR_POLL_WAITING
 } actor_state_e;
 
 typedef struct actor {
@@ -38,18 +34,17 @@ typedef struct actor {
 	actor_state_e state;
 	void *data;
 	void (*callback)(void * );
-	uint64_t scheduled_at;
-	uint64_t ttschedule;
+	time_t ttschedule;
 } actor_t;
 
-extern void actor_new(actor_t *thread, void (*callback)(void *), void * data);
+extern void actor_init(actor_t *thread, void (*callback)(void *), void * data);
 extern void actor_delete(actor_t *thread);
 extern void actor_schedule_head(actor_t *thread);
 extern void actor_schedule(actor_t *thread);
-extern void actor_timer(actor_t *thread, uint32_t timeout,
+extern void actor_timer(actor_t *thread, uint32_t delay_secs,
 			void (*callback)(void *), void *data);
-extern int actor_timer_mod(actor_t *thread, uint32_t new_timeout, void *data);
+extern void actor_timer_mod(actor_t *thread, uint32_t new_delay_secs,
+			    void *data);
 extern void actor_poll(void);
-extern void actor_init(void);
 
 #endif /* ACTOR_H */
