@@ -26,6 +26,8 @@
 #include <sys/poll.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/signalfd.h>
+#include <unistd.h>
 
 #include "mgmt_ipc.h"
 #include "iscsi_ipc.h"
@@ -35,8 +37,11 @@
 #include "iscsi_ipc.h"
 #include "actor.h"
 #include "initiator.h"
+#include "iscsi_err.h"
 
-static int reap_count;
+static unsigned int reap_count;
+
+#define REAP_WAKEUP 1000 /* in millisecs */
 
 void reap_inc(void)
 {
@@ -49,7 +54,7 @@ void reap_proc(void)
 
 	/*
 	 * We don't really need reap_count, but calling wait() all the
-	 * time seems execessive.
+	 * time seems excessive.
 	 */
 	max_reaps = reap_count;
 	for (i = 0; i < max_reaps; i++) {
@@ -79,7 +84,7 @@ int shutdown_callback(pid_t pid)
 
 	INIT_LIST_HEAD(&cb->list);
 	cb->pid = pid;
-	log_debug(1, "adding %d for shutdown cb\n", pid);
+	log_debug(1, "adding %d for shutdown cb", pid);
 	list_add_tail(&cb->list, &shutdown_callbacks);
 	return 0;
 }
@@ -89,7 +94,7 @@ static void shutdown_notify_pids(void)
 	struct shutdown_callback *cb;
 
 	list_for_each_entry(cb, &shutdown_callbacks, list) {
-		log_debug(1, "Killing %d\n", cb->pid);
+		log_debug(1, "Killing %d", cb->pid);
 		kill(cb->pid, SIGTERM);
 	}
 }
@@ -104,7 +109,7 @@ static int shutdown_wait_pids(void)
 		 * sign that it is gone.
 		 */
 		if (waitpid(cb->pid, NULL, WNOHANG)) {
-			log_debug(1, "%d done\n", cb->pid);
+			log_debug(1, "%d done", cb->pid);
 			list_del(&cb->list);
 			free(cb);
 		}
@@ -115,12 +120,12 @@ static int shutdown_wait_pids(void)
 
 #define POLL_CTRL	0
 #define POLL_IPC	1
-#define POLL_MAX	2
+#define POLL_ALARM	2
+#define POLL_MAX	3
 
 static int event_loop_stop;
 static queue_task_t *shutdown_qtask; 
 
-
 void event_loop_exit(queue_task_t *qtask)
 {
 	shutdown_qtask = qtask;
@@ -131,11 +136,26 @@ void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
 {
 	struct pollfd poll_array[POLL_MAX];
 	int res, has_shutdown_children = 0;
+	sigset_t sigset;
+	int sig_fd;
+
+	/* Mask off SIGALRM so we can recv it via signalfd */
+	sigemptyset(&sigset);
+	sigaddset(&sigset, SIGALRM);
+	sigprocmask(SIG_SETMASK, &sigset, NULL);
+
+	sig_fd = signalfd(-1, &sigset, SFD_NONBLOCK);
+	if (sig_fd == -1) {
+		log_error("signalfd failed: %m");
+		return;
+	}
 
 	poll_array[POLL_CTRL].fd = control_fd;
 	poll_array[POLL_CTRL].events = POLLIN;
 	poll_array[POLL_IPC].fd = mgmt_ipc_fd;
 	poll_array[POLL_IPC].events = POLLIN;
+	poll_array[POLL_ALARM].fd = sig_fd;
+	poll_array[POLL_ALARM].events = POLLIN;
 
 	event_loop_stop = 0;
 	while (1) {
@@ -148,7 +168,11 @@ void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
 				break;
 		}
 
-		res = poll(poll_array, POLL_MAX, ACTOR_RESOLUTION);
+		/* Runs actors and may set alarm for future actors */
+		actor_poll();
+
+		res = poll(poll_array, POLL_MAX, reap_count ? REAP_WAKEUP : -1);
+
 		if (res > 0) {
 			log_debug(6, "poll result %d", res);
 			if (poll_array[POLL_CTRL].revents)
@@ -156,6 +180,18 @@ void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
 
 			if (poll_array[POLL_IPC].revents)
 				mgmt_ipc_handle(mgmt_ipc_fd);
+
+			if (poll_array[POLL_ALARM].revents) {
+				struct signalfd_siginfo si;
+
+				if (read(sig_fd, &si, sizeof(si)) == -1) {
+					log_error("got sigfd read() error, errno (%d), "
+						  "exiting", errno);
+					break;
+				} else {
+					log_debug(1, "Poll was woken by an alarm");
+				}
+			}
 		} else if (res < 0) {
 			if (errno == EINTR) {
 				log_debug(1, "event_loop interrupted");
@@ -164,15 +200,20 @@ void event_loop(struct iscsi_ipc *ipc, int control_fd, int mgmt_ipc_fd)
 					  "exiting", res, errno);
 				break;
 			}
-		} else
-			actor_poll();
+		}
+
 		reap_proc();
+
 		/*
 		 * flush sysfs cache since kernel objs may
 		 * have changed as a result of handling op
 		 */
 		sysfs_cleanup();
 	}
+
 	if (shutdown_qtask)
-		mgmt_ipc_write_rsp(shutdown_qtask, MGMT_IPC_OK);
+		mgmt_ipc_write_rsp(shutdown_qtask, ISCSI_SUCCESS);
+
+	close(sig_fd);
+	sigprocmask(SIG_UNBLOCK, &sigset, NULL);
 }
