@@ -23,6 +23,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
+#include <dirent.h>
+#include <libmount/libmount.h>
 
 #include "initiator.h"
 #include "transport.h"
@@ -51,21 +53,9 @@ struct iscsi_session *session_find_by_sid(uint32_t sid)
 	return NULL;
 }
 
-/*
- * calculate parameter's padding
- */
-static unsigned int
-__padding(unsigned int param)
+const static unsigned int align_32_down(unsigned int param)
 {
-	int pad;
-
-	pad = param & 3;
-	if (pad) {
-		pad = 4 - pad;
-		log_debug(1, "parameter's value %d padded to %d bytes\n",
-			   param, param + pad);
-	}
-	return param + pad;
+	return param & ~0x3;
 }
 
 int iscsi_setup_authentication(struct iscsi_session *session,
@@ -77,7 +67,7 @@ int iscsi_setup_authentication(struct iscsi_session *session,
 	if (auth_cfg->username_in[0] || auth_cfg->password_in_length) {
 		/* sanity check the config */
 		if (auth_cfg->password_length == 0) {
-			log_warning("CHAP configuratoin has incoming "
+			log_warning("CHAP configuration has incoming "
 				    "authentication credentials but has no "
 				    "outgoing credentials configured.");
 			return EINVAL;
@@ -151,11 +141,11 @@ iscsi_copy_operational_params(struct iscsi_conn *conn,
 	conn->datadgst_en = conn_conf->DataDigest;
 
 	conn->max_recv_dlength =
-			__padding(conn_conf->MaxRecvDataSegmentLength);
+			align_32_down(conn_conf->MaxRecvDataSegmentLength);
 	if (conn->max_recv_dlength < ISCSI_MIN_MAX_RECV_SEG_LEN ||
 	    conn->max_recv_dlength > ISCSI_MAX_MAX_RECV_SEG_LEN) {
 		log_error("Invalid iscsi.MaxRecvDataSegmentLength. Must be "
-			 "within %u and %u. Setting to %u\n",
+			 "within %u and %u. Setting to %u",
 			  ISCSI_MIN_MAX_RECV_SEG_LEN,
 			  ISCSI_MAX_MAX_RECV_SEG_LEN,
 			  DEF_INI_MAX_RECV_SEG_LEN);
@@ -166,13 +156,13 @@ iscsi_copy_operational_params(struct iscsi_conn *conn,
 
 	/* zero indicates to use the target's value */
 	conn->max_xmit_dlength =
-			__padding(conn_conf->MaxXmitDataSegmentLength);
+			align_32_down(conn_conf->MaxXmitDataSegmentLength);
 	if (conn->max_xmit_dlength == 0)
 		conn->max_xmit_dlength = ISCSI_DEF_MAX_RECV_SEG_LEN;
 	if (conn->max_xmit_dlength < ISCSI_MIN_MAX_RECV_SEG_LEN ||
 	    conn->max_xmit_dlength > ISCSI_MAX_MAX_RECV_SEG_LEN) {
 		log_error("Invalid iscsi.MaxXmitDataSegmentLength. Must be "
-			 "within %u and %u. Setting to %u\n",
+			 "within %u and %u. Setting to %u",
 			  ISCSI_MIN_MAX_RECV_SEG_LEN,
 			  ISCSI_MAX_MAX_RECV_SEG_LEN,
 			  DEF_INI_MAX_RECV_SEG_LEN);
@@ -184,7 +174,7 @@ iscsi_copy_operational_params(struct iscsi_conn *conn,
 	/* session's operational parameters */
 	session->initial_r2t_en = session_conf->InitialR2T;
 	session->imm_data_en = session_conf->ImmediateData;
-	session->first_burst = __padding(session_conf->FirstBurstLength);
+	session->first_burst = align_32_down(session_conf->FirstBurstLength);
 	/*
 	 * some targets like netapp fail the login if sent bad first_burst
 	 * and max_burst lens, even when immediate data=no and
@@ -193,7 +183,7 @@ iscsi_copy_operational_params(struct iscsi_conn *conn,
 	if (session->first_burst < ISCSI_MIN_FIRST_BURST_LEN ||
 	    session->first_burst > ISCSI_MAX_FIRST_BURST_LEN) {
 		log_error("Invalid iscsi.FirstBurstLength of %u. Must be "
-			 "within %u and %u. Setting to %u\n",
+			 "within %u and %u. Setting to %u",
 			  session->first_burst,
 			  ISCSI_MIN_FIRST_BURST_LEN,
 			  ISCSI_MAX_FIRST_BURST_LEN,
@@ -202,11 +192,11 @@ iscsi_copy_operational_params(struct iscsi_conn *conn,
 		session->first_burst = DEF_INI_FIRST_BURST_LEN;
 	}
 
-	session->max_burst = __padding(session_conf->MaxBurstLength);
+	session->max_burst = align_32_down(session_conf->MaxBurstLength);
 	if (session->max_burst < ISCSI_MIN_MAX_BURST_LEN ||
 	    session->max_burst > ISCSI_MAX_MAX_BURST_LEN) {
 		log_error("Invalid iscsi.MaxBurstLength of %u. Must be "
-			  "within %u and %u. Setting to %u\n",
+			  "within %u and %u. Setting to %u",
 			   session->max_burst, ISCSI_MIN_MAX_BURST_LEN,
 			   ISCSI_MAX_MAX_BURST_LEN, DEF_INI_MAX_BURST_LEN);
 		session_conf->MaxBurstLength = DEF_INI_MAX_BURST_LEN;
@@ -215,7 +205,7 @@ iscsi_copy_operational_params(struct iscsi_conn *conn,
 
 	if (session->first_burst > session->max_burst) {
 		log_error("Invalid iscsi.FirstBurstLength of %u. Must be "
-			  "less than iscsi.MaxBurstLength. Setting to %u\n",
+			  "less than iscsi.MaxBurstLength. Setting to %u",
 			   session->first_burst, session->max_burst);
 		session_conf->FirstBurstLength = session->max_burst;
 		session->first_burst = session->max_burst;
@@ -261,7 +251,7 @@ int iscsi_setup_portal(struct iscsi_conn *conn, char *address, int port)
 	return 0;
 }
 
-int host_set_param(struct iscsi_transport *t,
+static int host_set_param(struct iscsi_transport *t,
 		   uint32_t host_no, int param, char *value,
 		   int type)
 {
@@ -273,7 +263,7 @@ int host_set_param(struct iscsi_transport *t,
 		log_error("can't set operational parameter %d for "
 			  "host %d, retcode %d (%d)", param, host_no,
 			  rc, errno);
-		return rc;
+		return ISCSI_ERR_INVAL;
 	}
 	return 0;
 }
@@ -324,12 +314,32 @@ int iscsi_host_set_params(struct iscsi_session *session)
 	return 0;
 }
 
-#define MAX_SESSION_PARAMS 32
+static inline void iscsi_session_clear_param(struct iscsi_session *session,
+					     int param)
+{
+	session->param_mask &= ~(1ULL << param);
+}
 
-int iscsi_session_set_params(struct iscsi_conn *conn)
+void iscsi_session_init_params(struct iscsi_session *session)
+{
+	session->param_mask = ~0ULL;
+	if (!(session->t->caps & CAP_MULTI_R2T))
+		iscsi_session_clear_param(session, ISCSI_PARAM_MAX_R2T);
+	if (!(session->t->caps & CAP_HDRDGST))
+		iscsi_session_clear_param(session, ISCSI_PARAM_HDRDGST_EN); 
+	if (!(session->t->caps & CAP_DATADGST))
+		iscsi_session_clear_param(session, ISCSI_PARAM_DATADGST_EN); 
+	if (!(session->t->caps & CAP_MARKERS)) {
+		iscsi_session_clear_param(session, ISCSI_PARAM_IFMARKER_EN);
+		iscsi_session_clear_param(session, ISCSI_PARAM_OFMARKER_EN);
+	}
+}
+
+#define MAX_SESSION_NEG_PARAMS 16
+
+int iscsi_session_set_neg_params(struct iscsi_conn *conn)
 {
 	struct iscsi_session *session = conn->session;
-	struct iscsi_transport *t = session->t;
 	int i, rc;
 	uint32_t one = 1, zero = 0;
 	struct connparam {
@@ -337,7 +347,7 @@ int iscsi_session_set_params(struct iscsi_conn *conn)
 		int type;
 		void *value;
 		int conn_only;
-	} conntbl[MAX_SESSION_PARAMS] = {
+	} conntbl[MAX_SESSION_NEG_PARAMS] = {
 		{
 			.param = ISCSI_PARAM_MAX_RECV_DLENGTH,
 			.value = &conn->max_recv_dlength,
@@ -411,18 +421,61 @@ int iscsi_session_set_params(struct iscsi_conn *conn)
 		}, {
 			.param = ISCSI_PARAM_EXP_STATSN,
 			.value = &conn->exp_statsn,
-			.type = ISCSI_INT,
+			.type = ISCSI_UINT,
 			.conn_only = 1,
-		}, {
-			.param = ISCSI_PARAM_TARGET_NAME,
-			.conn_only = 0,
-			.type = ISCSI_STRING,
-			.value = session->target_name,
 		}, {
 			.param = ISCSI_PARAM_TPGT,
 			.value = &session->portal_group_tag,
 			.type = ISCSI_INT,
 			.conn_only = 0,
+		},
+	};
+
+	iscsi_session_init_params(session);
+
+	/* Entered full-feature phase! */
+	for (i = 0; i < MAX_SESSION_NEG_PARAMS; i++) {
+		if (conn->id != 0 && !conntbl[i].conn_only)
+			continue;
+
+		if (!(session->param_mask & (1ULL << conntbl[i].param)))
+			continue;
+
+		rc = ipc->set_param(session->t->handle, session->id,
+				   conn->id, conntbl[i].param, conntbl[i].value,
+				   conntbl[i].type);
+		if (rc && rc != -ENOSYS) {
+			log_error("can't set operational parameter %d for "
+				  "connection %d:%d, retcode %d (%d)",
+				  conntbl[i].param, session->id, conn->id,
+				  rc, errno);
+			return EPERM;
+		}
+
+		print_param_value(conntbl[i].param, conntbl[i].value,
+				  conntbl[i].type);
+	}
+
+	return 0;
+}
+
+#define MAX_SESSION_PARAMS 20
+
+int iscsi_session_set_params(struct iscsi_conn *conn)
+{
+	struct iscsi_session *session = conn->session;
+	int i, rc;
+	struct connparam {
+		int param;
+		int type;
+		void *value;
+		int conn_only;
+	} conntbl[MAX_SESSION_PARAMS] = {
+		{
+			.param = ISCSI_PARAM_TARGET_NAME,
+			.conn_only = 0,
+			.type = ISCSI_STRING,
+			.value = session->target_name,
 		}, {
 			.param = ISCSI_PARAM_PERSISTENT_ADDRESS,
 			.value = session->nrec.conn[conn->id].address,
@@ -492,29 +545,41 @@ int iscsi_session_set_params(struct iscsi_conn *conn)
 			.param = ISCSI_PARAM_IFACE_NAME,
 			.value = session->nrec.iface.name,
 			.type = ISCSI_STRING,
+			.conn_only = 0,
 		}, {
 			.param = ISCSI_PARAM_INITIATOR_NAME,
 			.value = session->initiator_name,
 			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_BOOT_ROOT,
+			.value = session->nrec.session.boot_root,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_BOOT_NIC,
+			.value = session->nrec.session.boot_nic,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_BOOT_TARGET,
+			.value = session->nrec.session.boot_target,
+			.type = ISCSI_STRING,
+			.conn_only = 0,
+		}, {
+			.param = ISCSI_PARAM_DISCOVERY_SESS,
+			.value = &session->type,
+			.type = ISCSI_INT,
+			.conn_only = 0,
 		},
 	};
 
-	session->param_mask = ~0ULL;
-	if (!(t->caps & CAP_MULTI_R2T))
-		session->param_mask &= ~ISCSI_MAX_R2T;
-	if (!(t->caps & CAP_HDRDGST))
-		session->param_mask &= ~ISCSI_HDRDGST_EN;
-	if (!(t->caps & CAP_DATADGST))
-		session->param_mask &= ~ISCSI_DATADGST_EN;
-	if (!(t->caps & CAP_MARKERS)) {
-		session->param_mask &= ~ISCSI_IFMARKER_EN;
-		session->param_mask &= ~ISCSI_OFMARKER_EN;
-	}
+	iscsi_session_init_params(session);
 
 	/* some llds will send nops internally */
 	if (!iscsi_sysfs_session_supports_nop(session->id)) {
-		session->param_mask &= ~ISCSI_PING_TMO;
-		session->param_mask &= ~ISCSI_RECV_TMO;
+		iscsi_session_clear_param(session, ISCSI_PARAM_PING_TMO); 
+		iscsi_session_clear_param(session, ISCSI_PARAM_RECV_TMO);
 	}
 
 	/* Entered full-feature phase! */
@@ -562,6 +627,36 @@ TODO handle this
 	return 0;
 }
 
+int iscsi_set_net_config(struct iscsi_transport *t, iscsi_session_t *session,
+			 struct iface_rec *iface)
+{
+	if (t->template->set_net_config) {
+		/* uip needs the netdev name */
+		struct host_info hinfo;
+		int hostno, rc;
+
+		/* this assumes that the netdev or hw address is going to be
+		   set */
+		hostno = iscsi_sysfs_get_host_no_from_hwinfo(iface, &rc);
+		if (rc) {
+			log_debug(4, "Couldn't get host no.");
+			return rc;
+		}
+
+		/* uip needs the netdev name */
+		if (!strlen(iface->netdev)) {
+			memset(&hinfo, 0, sizeof(hinfo));
+			hinfo.host_no = hostno;
+			iscsi_sysfs_get_hostinfo_by_host_no(&hinfo);
+			strcpy(iface->netdev, hinfo.iface.netdev);
+		}
+
+		return t->template->set_net_config(t, iface, session);
+	}
+
+	return 0;
+}
+
 int iscsi_host_set_net_params(struct iface_rec *iface,
 			      struct iscsi_session *session)
 {
@@ -571,7 +666,7 @@ int iscsi_host_set_net_params(struct iface_rec *iface,
 	struct host_info hinfo;
 
 	log_debug(3, "setting iface %s, dev %s, set ip %s, hw %s, "
-		  "transport %s.\n",
+		  "transport %s.",
 		  iface->name, iface->netdev, iface->ipaddress,
 		  iface->hwaddress, iface->transport_name);
 
@@ -580,9 +675,18 @@ int iscsi_host_set_net_params(struct iface_rec *iface,
 
 	/* if we need to set the ip addr then set all the iface net settings */
 	if (!iface_is_bound_by_ipaddr(iface)) {
-		log_warning("Please set the iface.ipaddress for iface %s, "
-			    "then retry the login command.\n", iface->name);
-		return EINVAL;
+		if (t->template->set_host_ip == SET_HOST_IP_REQ) {
+			log_warning("Please set the iface.ipaddress for iface "
+				    "%s, then retry the login command.",
+				    iface->name);
+			return ISCSI_ERR_INVAL;
+		} else if (t->template->set_host_ip == SET_HOST_IP_OPT) {
+			log_info("Optional iface.ipaddress for iface %s "
+				 "not set.", iface->name);
+			return 0;
+		} else {
+			return ISCSI_ERR_INVAL;
+		}
 	}
 
 	/* these type of drivers need the netdev upd */
@@ -600,6 +704,10 @@ int iscsi_host_set_net_params(struct iface_rec *iface,
 		log_warning("Could not brining up netdev %s. Try running "
 			    "'ifup %s' first if login fails.", netdev, netdev);
 
+	rc = iscsi_set_net_config(t, session, iface);
+	if (rc != 0)
+		return rc;
+
 	rc = host_set_param(t, session->hostno,
 			    ISCSI_HOST_PARAM_IPADDRESS,
 			    iface->ipaddress, ISCSI_STRING);
