@@ -9,12 +9,29 @@
 
 struct list;
 
+struct session_timeout {
+	int abort_tmo;
+	int lu_reset_tmo;
+	int recovery_tmo;
+	int tgt_reset_tmo;
+};
+
+struct session_CHAP {
+	char username[AUTH_STR_MAX_LEN];
+	char password[AUTH_STR_MAX_LEN];
+	char username_in[AUTH_STR_MAX_LEN];
+	char password_in[AUTH_STR_MAX_LEN];
+};
+
 struct session_info {
 	struct list_head list;
 	/* local info */
 	struct iface_rec iface;
 	int sid;
 
+	struct session_timeout tmo;
+	struct session_CHAP chap;
+
 	/* remote info */
 	char targetname[TARGET_NAME_MAXLEN + 1];
 	int tpgt;
@@ -37,11 +54,14 @@ struct session_link_info {
 #define SESSION_INFO_ISCSI_STATE	0x4
 #define SESSION_INFO_SCSI_DEVS		0x8
 #define SESSION_INFO_HOST_DEVS		0x10
+#define SESSION_INFO_ISCSI_TIM          0x20
+#define SESSION_INFO_ISCSI_AUTH         0x40
 
 extern int session_info_create_list(void *data, struct session_info *info);
 extern void session_info_free_list(struct list_head *list);
-extern int session_info_print(int info_level, struct session_info *match_info);
+extern int session_info_print(int info_level, struct session_info *match_info,
+				int do_show);
 extern void session_info_print_tree(struct list_head *list, char *prefix,
-				    unsigned int flags);
+				    unsigned int flags, int do_show);
 
 #endif
